public abstract class DrawingTool
{

  public final int ID = -1;

  protected int w,h,x,y;

  protected float baseSize = 1.0;
  protected float minDist = 10.0;
  protected float maxDist = 10.0;

  protected float prevMouseX = 0.0;
  protected float prevMouseY = 0.0;
  protected float angle = 0.0;

  // transformed mouseX, mouseY
  protected float mouseTX, mouseTY;

  protected float currentDistance = 0.0; 

  protected boolean refreshFrame = false;
  protected boolean doClear = false;

  public int fillColor = 0xFFFFFFFF;
  public int strokeColor = 0xFFFFFFFF;

  public DrawingTool(int _x, int _y, int _w, int _h)
  {
    x = _x;
    y = _y;
    w = _w;
    h = _h;
  }

  public void clear()
  {
    doClear = true;
  }
  
  public void refresh()
  {
    refreshFrame = true;
  }


  // ------------------- drawing methods ---------------------

  // set and get return values because they may truncate value ranges

  private void beginDraw(PGraphics pg)
  {
    pg.beginDraw();
    pg.smooth();

    if (doClear)
    {
      pg.background(0x00000000);
      doClear = false;
    }

    if (mousePressed && isMouseInside())
    {

      transformMouse();

      currentDistance = dist( mouseTX, mouseTY, prevMouseX, prevMouseY);

      if ( currentDistance >= minDist)
      {
        angle = PI-atan2(mouseTX-prevMouseX, mouseTY-prevMouseY );

        refreshFrame = true;
      }
    }
  }

  private void endDraw(PGraphics pg) 
  {
    pg.dispose();
    pg.endDraw();
  }

  // this is where the magic happens!
  protected void _draw(PGraphics pg, float pressure) 
  {
  }


  public void draw(PGraphics pg) 
  {
    this.draw(pg, 1.0);
  }


  // if using a tablet
  public void draw(PGraphics pg, float pressure) 
  {
    this.beginDraw(pg);
    if (refreshFrame)
    {
      _draw(pg,pressure);
      refreshFrame = false;
      prevMouseX = mouseTX;
      prevMouseY = mouseTY;
    }
    this.endDraw(pg);
  }


  // --------------------- get/set methods --------------------

  public float setMinDist(float d)
  {
    minDist = max(0.0, d);
    return minDist;
  }

  public float getMinDist()
  {
    return minDist;
  }

  public float setMaxDist(float d)
  {
    maxDist = max(0.0, d);
    return maxDist;
  }

  public float getMaxDist()
  {
    return maxDist;
  }

  public float setBaseSize(float s)
  {
    baseSize = max(0.0, s);
    return baseSize;
  }
  public float getBaseSize(float s)
  {
    return baseSize;
  }

  public void setColor(int c)
  {
    this.fillColor = c;
  }


  // ---- other methods ----
  
  public void loadBrush(String brushFile)
  { }
  
  
  private boolean isMouseInside()
  {
    boolean inside = false;

    if (mouseX >= this.x && mouseX < (this.x+this.w) &&
      mouseY >= this.y && mouseY < (this.y+this.h)) inside = true;

    //println("inside / " + inside + " :: " + mouseX + "/" + this.x + ", " + mouseY + "/" + this.y);

    return inside;
  }

  private void transformMouse()
  {
    mouseTX = mouseX - this.x;
    mouseTY = mouseY - this.y;
  }
}
