/**
 * sound frame animator
 * by pixelpusher // evan raskob
 * http://pixelist.info
 * copyright 2009
 */

import fullscreen.*;
import japplemenubar.*;
import ddf.minim.*;
import codeanticode.protablet.*;
import controlP5.*;
import processing.opengl.*;

static final int LINE_TOOL = 0;
static final int CIRCLE_TOOL = 1;
static final int SQUARE_TOOL = 2;
static final int BRUSH_TOOL = 3;

static final int MODE_RANDOM = 0;
static final int MODE_SEQUENTIAL = 1;

Minim minim;
AudioSnippet snip;
ControlP5 controlP5;
FullScreen fs; 

DrawingTool currentTool;

DrawingTool tools[] = new DrawingTool[5];

Tablet tablet;

PGraphics currentFrame;
PFont font;
PImage prevFrame;

static final int fps = 10;
int chunkLength; 
int loopBegin;
int loopEnd;
int randChunk, chunks, numChunksShown=0;
int[] chunksShown;
int lastDonePressTime;

float currentSize = 7.0;

// go in order, or jump randomly?
int mode  = MODE_RANDOM;   
//int mode  = MODE_SEQUENTIAL;   

float currentToolColor = 255.0;
int toolColorX, toolColorY;
controlP5.Controller tc;

static final int FRAME_W = 720;
static final int FRAME_H = 480;

static final int ProgressBarHeight = 456;
static final int ProgressBarWidth  = 40;
static final int ProgressBarX = FRAME_W + 20;
static final int ProgressBarY = 40;
float pixelsPerChunk;

String outFileNamePrefix = "frame";
String outFileFolder;
String audioFileLoadPath;
int fileNameSequence = 0;
boolean debug = false;
int imgStartY = 16;


void setup()
{
  size(864,578);
  frameRate(60);
  controlP5 = new ControlP5(this);
  tablet = new Tablet(this); 

  prevFrame = createImage(FRAME_W, FRAME_H, RGB);
  
  // read in options from text file - ghetto, but easy
  
  
    
lastDonePressTime = 0;

  outFileFolder = selectFolder("Choose a folder to save frames into");

  if (outFileFolder == null)
  {
    println("Bad save folder!");
    exit();
  }

  String[] config = loadStrings(outFileFolder + "/config.txt");
  if (config != null)
  for (int i=0; i<config.length; i++)
  {
    String[] opts = split(config[i], "=");
    if (opts[0] == "CurrentFileSequence") fileNameSequence = int(opts[1]);
  } 

  font = loadFont("BitstreamVeraSansMono-Roman-12.vlw");
  textFont(font);

  tools[LINE_TOOL] = new LineTool(6,imgStartY,FRAME_W, FRAME_H);
  tools[CIRCLE_TOOL] = new CircleTool(6,imgStartY,FRAME_W, FRAME_H);
  tools[SQUARE_TOOL] = new SquareTool(6,imgStartY,FRAME_W, FRAME_H);
  tools[BRUSH_TOOL] = new BrushTool(6,imgStartY,FRAME_W, FRAME_H);

  tools[BRUSH_TOOL].loadBrush("brush01.svg");

  currentTool = tools[LINE_TOOL];

  currentFrame = createGraphics(FRAME_W, FRAME_H, P2D);
  currentFrame.smooth();

  PImage bmask = loadImage("small_button_mask.png");

  int xoff = 10;
  int yoff = imgStartY + 4 + FRAME_H;
  int bwidth = 54;
  int bpadding = 6;

  // line tool  ---------
  ControllerSprite sprite = new ControllerSprite(controlP5,loadImage("linebutton.png"),54,48);
  sprite.setMask(bmask);
  sprite.enableMask();

  controlP5.Button b = controlP5.addButton("lineBtn",1,xoff,yoff,54,48);
  b.setId(0);
  b.setSprite(sprite);

  // circle tool ---------
  sprite = new ControllerSprite(controlP5,loadImage("circlebutton.png"),54,48);
  sprite.setMask(bmask);
  sprite.enableMask();

  xoff += bwidth + bpadding;

  b = controlP5.addButton("circleBtn",1,xoff,yoff,54,48);
  b.setId(1);
  b.setSprite(sprite);  

  // square tool --------
  sprite = new ControllerSprite(controlP5,loadImage("sqbutton.png"),54,48);
  sprite.setMask(bmask);
  sprite.enableMask();

  xoff += bwidth + bpadding;

  b = controlP5.addButton("squareBtn",2,xoff,yoff,54,48);
  b.setId(2);
  b.setSprite(sprite);

  // brush tool --------
  sprite = new ControllerSprite(controlP5,loadImage("svgbrushbutton.png"),54,48);
  sprite.setMask(bmask);
  sprite.enableMask();

  xoff += bwidth + bpadding;

  b = controlP5.addButton("brushBtn",3,xoff,yoff,54,48);
  b.setId(3);
  b.setSprite(sprite);

  xoff += bwidth + bpadding;

  // tool draw size
  controlP5.Controller c = controlP5.addSlider("toolSize",1,101,20,xoff,yoff+3,100,20);
  c.setLabel("Size");
  c.setColorActive(0xFFFFFFFF);
  c.setColorBackground(0xFF333333);
  c.setColorForeground(0xFFEDEDED);
  c.setColorLabel(0xFFEDEDED);
  c.setColorValue(0xFF888888);

  // color slider
  tc = controlP5.addSlider("toolColor",0,255,255,xoff,yoff+29,100,20);
  tc.setLabel("Color");
  tc.setColorActive(0xFFFFFFFF);
  tc.setColorBackground(0xFF333333);
  tc.setColorForeground(0xFFEDEDED);
  tc.setColorLabel(0xFFEDEDED);
  tc.setColorValue(0xFF888888);

  toolColorX = xoff+100+40;
  toolColorY = yoff+29;

  //  xoff += c.getWidth() + bpadding;

  xoff += c.getWidth() + 80;

  c = controlP5.addBang("frameDone",xoff,yoff+3,60,15);
  c.setLabel("DONE!");
  c.setColorActive(0xFF33FF33);
  c.setColorBackground(0xFF888888);
  c.setColorForeground(0xFF11CC11);
  c.setColorLabel(0xFF44FF11);
  c.setColorValue(0xFF888888);

  xoff += c.getWidth() + 20;

  c = controlP5.addBang("clear",xoff,yoff+3,60,15);
  c.setLabel("CLEAR");
  c.setColorActive(0xFFFF0033);
  c.setColorBackground(0xFF333333);
  c.setColorForeground(0xFFCC0011);
  c.setColorLabel(0xFFEDEDED);
  c.setColorValue(0xFF888888);

  // --------- initialize audio ---------

  audioFileLoadPath = selectInput("Choose an audio file");  // Opens file chooser

  if (audioFileLoadPath == null) {
    // If a file was not selected
    println("No audio file was selected...");
    exit();
  } 
  else {
    // If a file was selected, print path to file
    println(audioFileLoadPath);
  }

  minim = new Minim(this);
  snip = minim.loadSnippet(audioFileLoadPath);
  snip.loop();

  chunkLength = int(1000.0/fps);

  if (debug) println("chunk length = " + chunkLength + "ms :: snip length=" + snip.length());


  // divide into chunkLength ms chunks
  // get random chunk
  // doesn't wrap around buffer or anything sophisticated like that...
  chunks = snip.length() / chunkLength;

  if (debug) println ("Prepared to render " + chunks + " chunks");

  pixelsPerChunk = ProgressBarHeight / float(chunks);

  // list of shown chunks
  chunksShown = new int[chunks];
  for (int i=0; i<chunks; i++)
    chunksShown[i] = 0;

  // look for finished frames

  File dir = new File(outFileFolder);

  File[] files = dir.listFiles(new FileRegExpFilter("frame" + nf(fileNameSequence,2) + "_.*\\.png"));
  int maxCount = files.length > 20000 ? 20000 : files.length;
  println("Found " +maxCount+ " frames already drawn");

  numChunksShown = maxCount;

  if (numChunksShown >= chunksShown.length) 
  {
    if (debug) println("DONE - do it again");
    numChunksShown = 0;
  }
  else
  {


    for (int i=0; i<files.length; i++)
    {

      if (!files[i].isDirectory()) 
      {
        String[] m = match(files[i].getName(), "frame"+nf(fileNameSequence,2)+"_(\\d+)");

        int frameIndex = int(m[1]);

        if (frameIndex < chunksShown.length)
        { 
          chunksShown[frameIndex] = 1;
          if (debug) println("Found chunk: " + frameIndex + ".");
        }
      }

      boolean gotChunk = false;

      while (!gotChunk && numChunksShown < chunksShown.length)
      {
        if (mode == MODE_RANDOM)
        {
          randChunk = floor(random(0,chunks));
        }
        else
          randChunk++;

        if (chunksShown[randChunk] == 0)
        {
          gotChunk = true;

          if (debug) println("Now drawing chunk:" + randChunk);
        }
      }
    }
  }



  if (debug) println ("Rand chunk[" + randChunk + "] at time - " + randChunk*chunkLength + "ms");

  snip.setLoopPoints(randChunk*chunkLength, (randChunk+1)*chunkLength);
  snip.cue(randChunk*chunkLength);

  // Create the fullscreen object
  fs = new FullScreen(this); 

  // enter fullscreen mode
  fs.enter(); 


  // clear screen
  currentTool.clear();

  currentTool.setBaseSize(currentSize);

  ellipseMode(CENTER);
  rectMode(CENTER);

  background(20);
  fill(0xFFFD9DFF);
  text("DRAWN TOGETHER by pixelpusher :: http://pixelist.info",6,12);

  fill(240);  
  text("Frames Animated:", ProgressBarX, ProgressBarY-24);
}



void draw()
{ 
  smooth();

  currentTool.draw(currentFrame, tablet.getPressure()+10.0);
  // currentTool.draw(currentFrame);
  imageMode(CORNER);
  if (mode == MODE_SEQUENTIAL)
  {
    tint(255,255);
    image(prevFrame,6,imgStartY, FRAME_W, FRAME_H);
    tint(255,225);
  }
  image(currentFrame,6,imgStartY, FRAME_W, FRAME_H);

  fill(20);
  rectMode(CORNER);
  noStroke();
  rect(0,FRAME_H+imgStartY,width,height-(FRAME_H+imgStartY));
  rect(ProgressBarX-6, ProgressBarY, ProgressBarWidth+48, ProgressBarHeight);

  // draw current color cube
  fill(currentToolColor);
  strokeWeight(1);
  stroke(0);
  rect(toolColorX, toolColorY, 20,20);

  // draw progress bar
  fill(10,30,10);
  stroke(0xFF27FF24);
  rect(ProgressBarX, ProgressBarY, ProgressBarWidth, ProgressBarHeight);

  noStroke();
  fill(0xFF27FF24);

  for (int i=0; i<chunks; i++)
  {
    if (chunksShown[i] == 1)
    {
      rect(ProgressBarX, ProgressBarY+i*pixelsPerChunk, ProgressBarWidth, max(1,pixelsPerChunk));
    }
  }  

  // draw current chunk
  stroke(200);
  strokeWeight(1);
  noFill();

  float currentChunkY = ProgressBarY+randChunk*pixelsPerChunk+0.5*pixelsPerChunk;

  rect(ProgressBarX, ProgressBarY+randChunk*pixelsPerChunk, ProgressBarWidth, pixelsPerChunk);
  line(ProgressBarX+ProgressBarWidth, currentChunkY, ProgressBarX+ProgressBarWidth+10, currentChunkY);
  fill(255.0 * ( (millis() / (float)chunkLength) % 1.0));
  rect(ProgressBarX+ProgressBarWidth+10, currentChunkY-5, 10, 10); 

  fill(255);
  if (debug)
    text("mx:" + mouseX +", my:"+ mouseY, 12,height-12);
  else
    text("Draw your interpretation of the sound and press DONE when finished.",12,height-12);

  fill(20);
  noStroke();
  rect(ProgressBarX, ProgressBarY-24, ProgressBarWidth+24, 24);
  fill(240);
  text(numChunksShown + " of " + chunks, ProgressBarX, ProgressBarY-12);
}


void getNextChunk()
{
  boolean gotChunk = false;

  // found unused chunk
  chunksShown[randChunk] = 1;

  currentFrame.save(outFileFolder + "/" + outFileNamePrefix + nf(fileNameSequence,2) + "_" + nf(randChunk,7) + ".png");
  prevFrame.copy(currentFrame, 0,0,FRAME_W,FRAME_H, 0,0,FRAME_W,FRAME_H);
  currentTool.clear();

  numChunksShown++;

  if (numChunksShown >= chunksShown.length) 
  {
    if (debug) println("DONE!");

    saveFramesToMovie();

    fileNameSequence++;
    numChunksShown = 0;
    for (int cc=0; cc<chunksShown.length; cc++)
      chunksShown[cc] = 0;
  }

  while (!gotChunk && numChunksShown < chunksShown.length)
  {
    if (mode == MODE_RANDOM)
    {
      randChunk = floor(random(0,chunks));
    }
    else
      randChunk++;

    if (chunksShown[randChunk] == 0)
    {
      gotChunk = true;

      if (debug) println("Now drawing chunk:" + randChunk);
    }
  }

  snip.setLoopPoints(randChunk*chunkLength, min(snip.length(), randChunk*chunkLength + 2*chunkLength));
  snip.cue(randChunk*chunkLength);
}





void stop()
{
  // always close Minim audio classes when you are done with them
  snip.close();
  minim.stop();

  // save current state
  String[] outStrings = {"CurrentFileSequence="+fileNameSequence}; 

  saveStrings(outFileFolder + "/config.txt", outStrings);

  super.stop();
  
}



void controlEvent(ControlEvent theEvent) {
  //println("got a control event from controller with id "+theEvent.controller().id() + "and label: " + theEvent.label());
  //println("Event name: " + theEvent.name());
  //println(theEvent.controller().stringValue());

  boolean toolChanged = false;

  switch(theEvent.controller().id()) {
    case(0):
    currentTool = tools[theEvent.controller().id()];
    toolChanged = true;
    break;
    case(1):
    currentTool = tools[theEvent.controller().id()];
    toolChanged = true;
    break;
    case(2):
    currentTool = tools[theEvent.controller().id()];
    toolChanged = true;        
    break; 
    case(3):
    currentTool = tools[theEvent.controller().id()];
    toolChanged = true;
    break;
  }

  if (toolChanged)
  {
    // maybe have a start() for a tool? and a finish()?
    //currentTool.finish();
    currentTool.setColor(color(currentToolColor));
    currentTool.setBaseSize(currentSize);
    //currentTool.start();
  }
}


void clear()
{
  //background(20);
  currentTool.clear();
}

void frameDone()
{
  int currentTime = millis();
  if (currentTime-lastDonePressTime > 800)
  {
    lastDonePressTime = currentTime;
    getNextChunk();
  }
}

void toolSize(float val)
{
  //println("got toolSize event:" + val);
  currentSize = val;
  currentTool.setBaseSize(val);
}


void toolColor(float val)
{
  color c = color(val);
  currentTool.setColor(c);
  // need this when we switch tools
  currentToolColor = val;

  tc.setColorBackground(c);
}


void keyReleased()
{
  if (key == 'd') debug = !debug;
}
